@@ -34,10 +34,11 @@ def run_forecast_ensemble(
         training_holdout_pct=0.3,
         frequency="week",
         period_format=None,
-        model_list=None):
+        model_list=None,
+        prediction_lag=0.0):
     # Initialize with copy of input date
+    dataframe[DATE_COLUMN_NAME] = pd.to_datetime(dataframe[date_column])
     df = dataframe.copy()
-    df[DATE_COLUMN_NAME] = pd.to_datetime(df[date_column])
 
     # Creates time series and ensures proper naming and frequency
     frequency = FREQUENCY_MAPPING.get(frequency)
@@ -51,6 +52,13 @@ def run_forecast_ensemble(
     # Create Future Forecast Periods
     start_date = pd.to_datetime(dataframe[DATE_COLUMN_NAME]).max() + relativedelta(days=1)
     end_date = start_date + relativedelta(years=forecast_horizon_years)
+    # Add additional period if lag is requested
+    if prediction_lag != 0:
+        if frequency == "W-MON":
+            delta = relativedelta(days=7)
+        elif frequency == "MS":
+            delta = relativedelta(months=1)
+        end_date = end_date + delta
     period_list = pd.date_range(start=start_date, end=end_date, freq=frequency)
 
     # Mark dataframe with training/testing split
@@ -124,16 +132,21 @@ def run_forecast_ensemble(
     ensemble_prediction_columns = [column for column in return_dataframe.columns if column.startswith("weighted_predicted_values")]
     ensemble_prediction_eval = " + ".join(ensemble_prediction_columns)
     return_dataframe["predicted_values_Ensemble"] = return_dataframe.eval(ensemble_prediction_eval)
+
     # Create ensemble weighted forecast
     ensemble_forecast_columns = [column for column in return_dataframe.columns if column.startswith("weighted_forecast_values")]
     ensemble_forecast_eval = " + ".join(ensemble_forecast_columns)
     return_dataframe["forecast_values_Ensemble"] = return_dataframe.eval(ensemble_forecast_eval)
+    forecast_columns = [column for column in return_dataframe.columns if column.startswith("forecast_values")]
+    if prediction_lag != 0:
+        return_dataframe = utils.apply_lag(return_dataframe, forecast_columns, prediction_lag)
+
     # Calculate ensemble metrics
     ensemble_train_df = return_dataframe[~return_dataframe["predicted_values_Ensemble"].isnull()]
     performance_metrics = utils.get_model_statistics(ensemble_train_df["predicted_values_Ensemble"], ensemble_train_df["y"])
     consolidated_metrics = utils.consolidate_scores(performance_metrics, ensemble_train_df["y"].mean())
+
     # Filter to only the required columns
-    forecast_columns = [column for column in return_dataframe.columns if column.startswith("forecast_values")]
     return_columns = ["y"] + forecast_columns
     return_dataframe = return_dataframe[return_columns]
 
@@ -152,11 +165,10 @@ def run_forecast_ensemble(
 
     ensemble.append({
         "name": "ensemble",
-        "model": None,
         "ensemble_dataframe": return_dataframe,
         "performance_metrics": performance_metrics,
         "consolidated_metrics": consolidated_metrics,
-        "weight": None
+        "degraded_accuracies": degraded_accuracies
     })
 
     return ensemble
