@@ -1,11 +1,12 @@
 # Class to extend for each model
 
 import pandas as pd
+import numpy as np
 from datetime import datetime
 from dateutil.relativedelta import relativedelta
 
 from fbprophet import Prophet
-from greykite.framework.templates.autogen.forecast_config import ForecastConfig, MetadataParam, EvaluationPeriodParam
+from greykite.framework.templates.autogen.forecast_config import ForecastConfig, MetadataParam, ModelComponentsParam, EvaluationPeriodParam
 from greykite.framework.templates.forecaster import Forecaster
 from greykite.framework.templates.model_templates import ModelTemplateEnum
 from xgboost import XGBRegressor
@@ -39,13 +40,14 @@ class BaseModel:
 
     def evaluate(self, dataframe: pd.DataFrame, train_end_date: datetime, frequency: str,
                  forecast_period_list: list) -> dict:
-        """Function to evaluate the perfomrance of the model against a particular dataset"""
+        """Function to evaluate the performance of the model against a particular dataset"""
         actual, predicted, forecast = self.run(
             dataframe=dataframe,
             train_end=train_end_date,
             frequency=frequency,
             forecast_period_list=forecast_period_list)
         result_package = utils.get_model_statistics(y=actual, yhat=predicted)
+        result_package.update(utils.consolidate_scores(result_package, actual.mean()))
         result_package["name"] = self.__name__
         result_package["forecast_dataframe"] = pd.DataFrame(
                 index=pd.Series(forecast.index),
@@ -89,6 +91,7 @@ class GreyKite(BaseModel):
         result = forecaster.run_forecast_config(  # result is also stored as `forecaster.forecast_result`.
             df=dataframe,
             config=ForecastConfig(
+                model_components_param=ModelComponentsParam(seasonality={"yearly_seasonality": 40}),
                 evaluation_period_param=EvaluationPeriodParam(test_horizon=test_horizon),
                 model_template=ModelTemplateEnum.SILVERKITE.name,
                 coverage=0.95,
@@ -125,8 +128,8 @@ class FBProphet(BaseModel):
         """
         df = dataframe.rename(columns={DATE_COLUMN_NAME: "ds", VALUE_COLUMN_NAME: "y"})
 
-        model = Prophet()
-        model.fit(df, iter=10)
+        model = Prophet(yearly_seasonality=20)
+        model.fit(df, iter=100)
 
         future = model.make_future_dataframe(periods=len(forecast_period_list), freq=frequency)
         forecast = model.predict(future)
@@ -203,6 +206,8 @@ class Naive(BaseModel):
 
         dataframe_test["predicted"] = predicted
         dataframe_test = dataframe_test.set_index("period")
+        dataframe_test = dataframe_test[np.isfinite(dataframe_test).all(axis=1)]
+
         dataframe = dataframe.set_index("period")
 
         return dataframe_test[VALUE_COLUMN_NAME], dataframe_test["predicted"], dataframe["yhat"]
@@ -211,125 +216,42 @@ class Naive(BaseModel):
 class XGBoost(BaseModel):
     OBJECTIVE = "reg:squarederror"
     N_ESTIMATORS = 100
-    TRAINING_PERIODS = 12
-    PREVIOUS_PERIODS = 6
+    PREVIOUS_PERIODS = 0
 
     def __init__(self):
         """Initializes xgboost model"""
         super().__init__()
         self.__name__ = "XGBoost"
 
-    def create_time_series_features(self, time_series_df):
-        result_df = pd.DataFrame(time_series_df[[VALUE_COLUMN_NAME]])
-        for ii in range(1, self.PREVIOUS_PERIODS + 1):
+    def create_time_series_features(self, time_series_df, frequency, backtest=False):
+        result_df = time_series_df.copy()
+        if frequency == "W-MON":
+            self.PREVIOUS_PERIODS = 53
+            result_df["year"] = result_df.index.year
+            result_df["period_number"] = result_df.index.isocalendar().week.astype(str)
+        else:
+            self.PREVIOUS_PERIODS = 12
+            result_df["period_number"] = result_df.index.month.astype(str)
+        previous_periods = self.PREVIOUS_PERIODS
+        if backtest:
+            if len(result_df) < (2 * self.PREVIOUS_PERIODS):
+                previous_periods = int(len(result_df) / 2)
+        for ii in range(1, previous_periods):
+            # Create recent lag periods
             result_df[f"{VALUE_COLUMN_NAME}_{ii}"] = time_series_df[VALUE_COLUMN_NAME].shift(ii)
         result_df.dropna(inplace=True)
+        period_indexes = pd.get_dummies(result_df.period_number)
+        dataframe = pd.concat([result_df, period_indexes], axis=1).drop(columns=["period_number"])
 
-        return result_df
-
-    def step_forecast(self, train_df, test_x, feature_columns):
-        # fit an xgboost model and make a one step prediction
-        # split into input and output columns
-        train_x = train_df[feature_columns]
-        train_y = train_df[VALUE_COLUMN_NAME]
-        # fit model
-        model = XGBRegressor(objective=self.OBJECTIVE, n_estimators=self.N_ESTIMATORS)
-        model.fit(train_x, train_y)
-        # make a one-step prediction
-        yhat = model.predict(test_x)
-
-        return yhat[0]
-
-    def full_forecast(self, train_df, feature_columns, forecast_period_list):
-        # fit an xgboost model and make a prediction for the full forecast_horizon
-        # split into input and output columns
-        train_x = train_df[feature_columns]
-        train_y = train_df[VALUE_COLUMN_NAME]
-
-        # fit model
-        model = XGBRegressor(objective=self.OBJECTIVE, n_estimators=self.N_ESTIMATORS)
-        model.fit(train_x, train_y)
-
-        cc = 0
-        previous_record = train_df.rename(columns={"y": "y_0"}).iloc[-1]
-        predictions = {}
-        for period in forecast_period_list:
-            if cc == 0:
-                test_record = previous_record.copy()
-            else:
-                previous_record = test_record.copy()
-            for ii in range(0, self.PREVIOUS_PERIODS):
-                test_record[f"y_{ii + 1}"] = previous_record[f"y_{ii}"]
-            test_x = pd.DataFrame(test_record[feature_columns]).T
-
-            # Run Forecast
-            prediction = model.predict(test_x)
-            test_record["y_0"] = prediction
-            predictions[period] = prediction[0]
-
-            cc += 1
-        predictions = pd.Series(predictions)
-
-        return predictions
-
-    def run(self, dataframe: pd.DataFrame, train_end: datetime, frequency: str,
-            forecast_period_list: pd.DataFrame) -> tuple:
-        dataframe = dataframe.copy()
-        # Prepare dataframe
-        dataframe.set_index("period", inplace=True)
-        dataframe = dataframe.asfreq(frequency)
-        dataframe = self.create_time_series_features(dataframe)
-        predicted = []
-        # Split into train and test
-        train_mask = dataframe.index <= train_end
-        train = dataframe[train_mask]
-        test = dataframe[~train_mask]
-        feature_columns = list(set(dataframe.columns) - set(VALUE_COLUMN_NAME))
-        # Step over each time period and get result
-        for index, row in test.iterrows():
-            # split test row into input and output columns
-            x, y = pd.DataFrame(row[feature_columns]).T, row[VALUE_COLUMN_NAME]
-
-            # fit model on history and make a prediction
-            yhat = self.step_forecast(train, x, feature_columns)
-
-            # store forecast in list of predictions
-            predicted.append(yhat)
-
-            # add actual observation to history for the next loop
-            train = pd.concat([train, pd.DataFrame(row).T])
-
-        # Run full forecast
-        forecast_predictions = self.full_forecast(dataframe, feature_columns, forecast_period_list)
-
-        return test[VALUE_COLUMN_NAME], predicted, forecast_predictions
-
-
-class RandomForest(BaseModel):
-    N_ESTIMATORS = 100
-    TRAINING_PERIODS = 12
-    PREVIOUS_PERIODS = 6
-
-    def __init__(self):
-        """Initializes random forest model"""
-        super().__init__()
-        self.__name__ = "RandomForest"
-
-    def create_time_series_features(self, time_series_df):
-        result_df = pd.DataFrame(time_series_df[[VALUE_COLUMN_NAME]])
-        for ii in range(1, self.PREVIOUS_PERIODS + 1):
-            result_df[f"{VALUE_COLUMN_NAME}_{ii}"] = time_series_df[VALUE_COLUMN_NAME].shift(ii)
-        result_df.dropna(inplace=True)
-
-        return result_df
+        return dataframe
 
     def step_forecast(self, train_df: pd.DataFrame, test_x: pd.DataFrame, feature_columns: list):
-        # fit a rf model and make a one step prediction
+        # fit a xgboost model and make a one step prediction
         # split into input and output columns
         train_x = train_df[feature_columns]
         train_y = train_df[VALUE_COLUMN_NAME]
         # fit model
-        model = RandomForestRegressor(n_estimators=self.N_ESTIMATORS)
+        model = XGBRegressor(objective=self.OBJECTIVE, n_estimators=self.N_ESTIMATORS)
         model.fit(train_x, train_y)
 
         # make a one-step prediction
@@ -337,26 +259,45 @@ class RandomForest(BaseModel):
 
         return yhat[0]
 
-    def full_forecast(self, train_df, feature_columns, forecast_period_list):
+    def full_forecast(self, train_df, feature_columns, forecast_period_list, frequency):
         # fit an xgboost model and make a prediction for the full forecast_horizon
         # split into input and output columns
         train_x = train_df[feature_columns]
         train_y = train_df[VALUE_COLUMN_NAME]
 
         # fit model
-        model = RandomForestRegressor(n_estimators=self.N_ESTIMATORS)
+        model = XGBRegressor(objective=self.OBJECTIVE, n_estimators=self.N_ESTIMATORS)
         model.fit(train_x, train_y)
 
         cc = 0
         previous_record = train_df.rename(columns={"y": "y_0"}).iloc[-1]
         predictions = {}
         for period in forecast_period_list:
+            # Initialize previous period
             if cc == 0:
                 test_record = previous_record.copy()
+                # Update period IDs
+                period_id_columns = [col for col in previous_record.index.values if "_" not in col]
+                test_record[period_id_columns] = 0
+                if frequency == "W-MON":
+                    period_id = forecast_period_list[cc].isocalendar()[1]
+                else:
+                    period_id = forecast_period_list[cc].month
+                test_record[str(period_id)] = 1
             else:
+                # Copy previous record and update period_id
                 previous_record = test_record.copy()
+                # Update period IDs
+                test_record[str(period_id)] = 0
+                if frequency == "W-MON":
+                    period_id = forecast_period_list[cc].isocalendar()[1]
+                else:
+                    period_id = forecast_period_list[cc].month
+                test_record[str(period_id)] = 1
+            # Update previous values
             for ii in range(0, self.PREVIOUS_PERIODS):
                 test_record[f"y_{ii + 1}"] = previous_record[f"y_{ii}"]
+
             test_x = pd.DataFrame(test_record[feature_columns]).T
 
             # Run Forecast
@@ -372,18 +313,20 @@ class RandomForest(BaseModel):
     def run(self, dataframe: pd.DataFrame, train_end: datetime, frequency: str,
             forecast_period_list: pd.DataFrame) -> tuple:
         dataframe = dataframe.copy()
-
-        # Prepare dataframe
         dataframe.set_index("period", inplace=True)
         dataframe = dataframe.asfreq(frequency)
-        dataframe = self.create_time_series_features(dataframe)
+
+        # Prepare dataframe
+        dataframe_backtest = self.create_time_series_features(dataframe, frequency, backtest=True)
 
         predicted = []
         # Split into train and test
-        train_mask = dataframe.index <= train_end
-        train = dataframe[train_mask]
-        test = dataframe[~train_mask]
-        feature_columns = list(set(dataframe.columns) - set(VALUE_COLUMN_NAME))
+        if train_end < dataframe_backtest.index.min():
+            train_end = dataframe_backtest.index.min()
+        train_mask = dataframe_backtest.index <= train_end
+        train = dataframe_backtest[train_mask]
+        test = dataframe_backtest[~train_mask]
+        feature_columns = list(set(dataframe_backtest.columns) - set(VALUE_COLUMN_NAME))
 
         # Step over each time period and get result
         for index, row in test.iterrows():
@@ -400,13 +343,148 @@ class RandomForest(BaseModel):
             train = pd.concat([train, pd.DataFrame(row).T])
 
         # Run full forecast
-        forecast_predictions = self.full_forecast(dataframe, feature_columns, forecast_period_list)
+        dataframe_forecast = self.create_time_series_features(dataframe, frequency, backtest=False)
+        forecast_predictions = self.full_forecast(dataframe_forecast, feature_columns, forecast_period_list, frequency)
 
         return test[VALUE_COLUMN_NAME], predicted, forecast_predictions
 
 
+class RandomForest(BaseModel):
+    N_ESTIMATORS = 100
+    PREVIOUS_PERIODS = 0
+
+    def __init__(self):
+        """Initializes random forest model"""
+        super().__init__()
+        self.__name__ = "RandomForest"
+
+    def create_time_series_features(self, time_series_df, frequency, backtest=False):
+        result_df = time_series_df.copy()
+        if frequency == "W-MON":
+            self.PREVIOUS_PERIODS = 53
+            result_df["year"] = result_df.index.year
+            result_df["period_number"] = result_df.index.isocalendar().week.astype(str)
+        else:
+            self.PREVIOUS_PERIODS = 12
+            result_df["period_number"] = result_df.index.month.astype(str)
+        previous_periods = self.PREVIOUS_PERIODS
+        if backtest:
+            if len(result_df) < (2 * self.PREVIOUS_PERIODS):
+                previous_periods = int(len(result_df) / 2)
+        for ii in range(1, previous_periods):
+            # Create recent lag periods
+            result_df[f"{VALUE_COLUMN_NAME}_{ii}"] = time_series_df[VALUE_COLUMN_NAME].shift(ii)
+        result_df.dropna(inplace=True)
+        period_indexes = pd.get_dummies(result_df.period_number)
+        dataframe = pd.concat([result_df, period_indexes], axis=1).drop(columns=["period_number"])
+
+        return dataframe
+
+    def step_forecast(self, train_df: pd.DataFrame, test_x: pd.DataFrame, feature_columns: list):
+        # fit a rf model and make a one step prediction
+        # split into input and output columns
+        train_x = train_df[feature_columns]
+        train_y = train_df[VALUE_COLUMN_NAME]
+        # fit model
+        model = RandomForestRegressor(n_estimators=self.N_ESTIMATORS)
+        model.fit(train_x, train_y)
+
+        # make a one-step prediction
+        yhat = model.predict(test_x)
+
+        return yhat[0]
+
+    def full_forecast(self, train_df, feature_columns, forecast_period_list, frequency):
+        # fit an xgboost model and make a prediction for the full forecast_horizon
+        # split into input and output columns
+        train_x = train_df[feature_columns]
+        train_y = train_df[VALUE_COLUMN_NAME]
+
+        # fit model
+        model = RandomForestRegressor(n_estimators=self.N_ESTIMATORS)
+        model.fit(train_x, train_y)
+
+        cc = 0
+        previous_record = train_df.rename(columns={"y": "y_0"}).iloc[-1]
+        predictions = {}
+        for period in forecast_period_list:
+            # Initialize previous period
+            if cc == 0:
+                test_record = previous_record.copy()
+                # Update period IDs
+                period_id_columns = [col for col in previous_record.index.values if "_" not in col]
+                test_record[period_id_columns] = 0
+                if frequency == "W-MON":
+                    period_id = forecast_period_list[cc].isocalendar()[1]
+                else:
+                    period_id = forecast_period_list[cc].month
+                test_record[str(period_id)] = 1
+            else:
+                # Copy previous record and update period_id
+                previous_record = test_record.copy()
+                # Update period IDs
+                test_record[str(period_id)] = 0
+                if frequency == "W-MON":
+                    period_id = forecast_period_list[cc].isocalendar()[1]
+                else:
+                    period_id = forecast_period_list[cc].month
+                test_record[str(period_id)] = 1
+            # Update previous values
+            for ii in range(0, self.PREVIOUS_PERIODS):
+                test_record[f"y_{ii + 1}"] = previous_record[f"y_{ii}"]
+
+            test_x = pd.DataFrame(test_record[feature_columns]).T
+
+            # Run Forecast
+            prediction = model.predict(test_x)
+            test_record["y_0"] = prediction
+            predictions[period] = prediction[0]
+
+            cc += 1
+        predictions = pd.Series(predictions)
+
+        return predictions
+
+
+    def run(self, dataframe: pd.DataFrame, train_end: datetime, frequency: str,
+            forecast_period_list: pd.DataFrame) -> tuple:
+        dataframe = dataframe.copy()
+        dataframe.set_index("period", inplace=True)
+        dataframe = dataframe.asfreq(frequency)
+
+        # Prepare dataframe
+        dataframe_backtest = self.create_time_series_features(dataframe, frequency, backtest=True)
+
+        predicted = []
+        # Split into train and test
+        if train_end < dataframe_backtest.index.min():
+            train_end = dataframe_backtest.index.min()
+        train_mask = dataframe_backtest.index <= train_end
+        train = dataframe_backtest[train_mask]
+        test = dataframe_backtest[~train_mask]
+        feature_columns = list(set(dataframe_backtest.columns) - set(VALUE_COLUMN_NAME))
+
+        # Step over each time period and get result
+        for index, row in test.iterrows():
+            # split test row into input and output columns
+            X, y = pd.DataFrame(row[feature_columns]).T, row[VALUE_COLUMN_NAME]
+
+            # fit model on history and make a prediction
+            yhat = self.step_forecast(train, X, feature_columns)
+
+            # store forecast in list of predictions
+            predicted.append(yhat)
+
+            # add actual observation to history for the next loop
+            train = pd.concat([train, pd.DataFrame(row).T])
+
+        # Run full forecast
+        dataframe_forecast = self.create_time_series_features(dataframe, frequency, backtest=False)
+        forecast_predictions = self.full_forecast(dataframe_forecast, feature_columns, forecast_period_list, frequency)
+
+        return test[VALUE_COLUMN_NAME], predicted, forecast_predictions
+
 class SARIMA(BaseModel):
-    TRAINING_PERIODS = 12
 
     def __init__(self):
         """Initializes SARIMA model"""
@@ -419,34 +497,56 @@ class SARIMA(BaseModel):
         dataframe.set_index("period", inplace=True)
         dataframe = dataframe.asfreq(frequency)
         if frequency == "W-MON":
-            m = 52
-        else:
+            m = 52.5
+        elif frequency == "MS":
             m = 12
 
-        predicted = []
         # Split into train and test
         train_mask = dataframe.index <= train_end
         train = dataframe[train_mask]
         test = dataframe[~train_mask]
-        # Step over each time period and get result
-        for index, row in test.iterrows():
-            # fit model on history and make a prediction
+
+        # Create step-forward validation
+        predicted = pd.DataFrame()
+        previous_index = 0
+        period_horizon = len(forecast_period_list)
+        for ii in range(period_horizon, len(test), period_horizon):
             model = SARIMAX(train, order=(1, 1, 2), seasonal_order=(1, 1, 2, m))
             results = model.fit(method='powell')
 
-            # store forecast in list of predictions
-            predicted.append(results.forecast(steps=1))
+            # Identify subset of test to forecast against
+            test_subset = test.iloc[previous_index:ii]
+            previous_index = ii
 
-            # add actual observation to history for the next loop
-            train = pd.concat([train, pd.DataFrame(row).T])
+            # Create predictions and add to list
+            predictions = pd.DataFrame(index=test_subset.index, data=results.forecast(steps=period_horizon).values)
+            predicted = pd.concat([predicted, predictions])
 
-        model = SARIMAX(train, order=(1, 1, 2), seasonal_order=(1, 1, 2, m))
+            # Update training dataset
+            train = pd.concat([train, test_subset])
+
+            # Handle final test records
+            if (len(test) - ii) < period_horizon:
+                model = SARIMAX(train, order=(1, 1, 2), seasonal_order=(1, 1, 2, m))
+                results = model.fit(method='powell')
+
+                # Identify subset of test to forecast against
+                test_subset = test.iloc[ii:]
+                train = pd.concat([train, test_subset])
+
+                # Create predictions and add to list
+                predictions = pd.DataFrame(index=test_subset.index,
+                                           data=results.forecast(steps=len(test_subset)).values)
+                predicted = pd.concat([predicted, predictions])
+
+        # Create future forecast
+        model = SARIMAX(dataframe, order=(1, 1, 2), seasonal_order=(1, 1, 2, m))
         results = model.fit(method='powell')
         predictions = results.forecast(steps=len(forecast_period_list))
 
         forecast = pd.Series(index=forecast_period_list, data=predictions)
 
-        return test[VALUE_COLUMN_NAME], predicted, forecast
+        return test[VALUE_COLUMN_NAME], predicted[0].values, forecast
 
 
 def initialize_model(model_name: str):
